@@ -2,6 +2,8 @@
  * OS-specific definitions
  */
 
+#define ROOT_UID 0
+
 #ifdef __hpux
 #define DEFAULT_MAIL_COMMAND "/usr/bin/mailx"
 #define COMPRESS_COMMAND "/usr/contrib/bin/gzip"
@@ -16,6 +18,13 @@
 #define STATEFILE "/var/log/logrotate.status"
 #endif
 
+#ifdef __NetBSD__
+   #define DEFAULT_MAIL_COMMAND "/usr/bin/mail -s"
+   #define COMPRESS_COMMAND "/usr/bin/gzip"
+   #define UNCOMPRESS_COMMAND "/usr/bin/gunzip"
+   #define STATEFILE "/var/log/logrotate.status"
+#endif
+
 /*
  * Default settings for Linux - leave these last.
  */
